package level

import (
	"fmt"
	"image"
	"log"
	"math/rand"
	"path/filepath"
	"strings"
	"time"

	"github.com/jakecoffman/cp"

	"code.rocketnine.space/tslocum/doctorlectro/asset"
	"code.rocketnine.space/tslocum/doctorlectro/component"
	"code.rocketnine.space/tslocum/doctorlectro/world"
	"code.rocketnine.space/tslocum/gohan"
	"github.com/hajimehoshi/ebiten/v2"
	"github.com/lafriks/go-tiled"
)

func LoadMap() {
	filePath := "map/map.tmx"

	// Parse .tmx file.
	m, err := tiled.LoadFile(filePath, tiled.WithFileSystem(asset.FS))
	if err != nil {
		log.Fatalf("error parsing world: %+v", err)
	}

	// Load tileset.

	tileset := m.Tilesets[0]

	imgPath := filepath.Join("image/robot-tileset", tileset.Image.Source)
	f, err := asset.FS.Open(filepath.FromSlash(imgPath))
	if err != nil {
		panic(err)
	}
	defer f.Close()

	img, _, err := image.Decode(f)
	if err != nil {
		panic(err)
	}
	tilesetImg := ebiten.NewImageFromImage(img)

	// Load tiles.

	for i := uint32(0); i < uint32(tileset.TileCount); i++ {
		rect := tileset.GetTileRect(i)
		world.TileImages[i+tileset.FirstGID] = tilesetImg.SubImage(rect).(*ebiten.Image)
	}

	blueBeamTiles := []*ebiten.Image{
		world.TileImages[tileset.FirstGID+asset.TileBeamBlue1],
		world.TileImages[tileset.FirstGID+asset.TileBeamBlue2],
		world.TileImages[tileset.FirstGID+asset.TileBeamBlue3],
	}

	orangeBeamTiles := []*ebiten.Image{
		world.TileImages[tileset.FirstGID+asset.TileBeamOrange1],
		world.TileImages[tileset.FirstGID+asset.TileBeamOrange2],
		world.TileImages[tileset.FirstGID+asset.TileBeamOrange3],
	}

	isBeamTile := func(id uint32) bool {
		for _, beamID := range asset.BeamTiles {
			if id == beamID {
				return true
			}
		}
		return false
	}

	rectSegments := func(r image.Rectangle) [][4]int {
		return [][4]int{
			{r.Min.X, r.Min.Y, r.Max.X, r.Min.Y}, // Top
			{r.Min.X, r.Max.Y, r.Max.X, r.Max.Y}, // Bottom
			{r.Min.X, r.Min.Y, r.Min.X, r.Max.Y}, // Left
			{r.Max.X, r.Min.Y, r.Max.X, r.Max.Y}, // Right
		}
	}

	addCollisionRect := func(r image.Rectangle) []*cp.Shape {
		var shapes []*cp.Shape
		for _, seg := range rectSegments(r) {
			segment := cp.NewSegment(world.Space.StaticBody, cp.Vector{float64(seg[0]), float64(seg[1])}, cp.Vector{float64(seg[2]), float64(seg[3])}, 1)
			segment.SetElasticity(0)
			segment.SetFriction(0)

			shape := world.Space.AddShape(segment)
			shape.SetElasticity(0)
			shape.SetFriction(0)

			shapes = append(shapes, shape)
		}
		return shapes
	}

	animateTile := func(id uint32, sprite *component.Sprite) {
		if !isBeamTile(id) {
			return
		}

		blueBeam := id == asset.TileBeamBlue1 || id == asset.TileBeamBlue2 || id == asset.TileBeamBlue3
		if blueBeam {
			sprite.Frames = blueBeamTiles
		} else {
			sprite.Frames = orangeBeamTiles
		}
		sprite.NumFrames = len(sprite.Frames)
		sprite.Frame = rand.Intn(sprite.NumFrames)
		sprite.FrameTime = 100 * time.Millisecond
	}

	createTileEntity := func(t *tiled.LayerTile, x int, y int, decoration bool) gohan.Entity {
		tileX, tileY := TileToGameCoords(x, y)

		mapTile := gohan.NewEntity()
		mapTile.AddComponent(&component.Position{
			X: tileX,
			Y: tileY,
		})

		sprite := &component.Sprite{
			Image:          world.TileImages[t.Tileset.FirstGID+t.ID],
			HorizontalFlip: t.HorizontalFlip,
			VerticalFlip:   t.VerticalFlip,
		}
		if decoration {
			sprite.Layer = world.LayerDecorationBelow
		} else {
			sprite.Layer = world.LayerPlatform
		}

		animateTile(t.ID, sprite)

		mapTile.AddComponent(sprite)

		return mapTile
	}

	metalTiles := make([][]bool, m.Height)
	for y := range metalTiles {
		metalTiles[y] = make([]bool, m.Width)
	}

	var t *tiled.LayerTile
	for _, layer := range m.Layers {
		isDecoration := layer.Name == "Decorations"
		isDestructible := layer.Name == "Destructibles"
		for y := 0; y < m.Height; y++ {
			for x := 0; x < m.Width; x++ {
				t = layer.Tiles[y*m.Width+x]
				if t == nil || t.Nil {
					continue // No tile at this position.
				}

				tileImg := world.TileImages[t.Tileset.FirstGID+t.ID]
				if tileImg == nil {
					continue
				}
				e := createTileEntity(t, x, y, isDecoration)

				if isDestructible {
					r := image.Rect(x*16, y*16, (x+1)*16, (y+1)*16)
					shapes := addCollisionRect(r)
					world.DestructibleRects = append(world.DestructibleRects, r)
					world.DestructibleEntities = append(world.DestructibleEntities, e)
					world.DestructibleShapes = append(world.DestructibleShapes, shapes)
				} else if !isDecoration && !isBeamTile(t.ID) {
					metalTiles[y][x] = true
				}
			}
		}
	}

	// Load objects.

	createTriggerEntity := func(obj *tiled.Object) gohan.Entity {
		var action world.TriggerAction
		switch obj.Name {
		case "GETJUMP":
			action = world.TriggerJump
		case "GETMAGNET":
			action = world.TriggerMagnetize
		case "GETFIRE":
			action = world.TriggerFire
		case "GETSWITCH":
			action = world.TriggerSwitch
		case "GETDATA":
			action = world.TriggerData
		default:
			panic(fmt.Sprintf("unknown trigger %s", obj.Name))
		}

		mapTile := gohan.NewEntity()
		mapTile.AddComponent(&component.Position{
			X: obj.X,
			Y: obj.Y - 16,
		})

		sprite := &component.Sprite{
			Image: world.TileImages[obj.GID],
			Layer: world.LayerDecorationBelow,
		}
		mapTile.AddComponent(sprite)

		r := ObjectToRect(obj)
		world.TriggerRects = append(world.TriggerRects, r)
		world.TriggerActions = append(world.TriggerActions, action)
		world.TriggerEntities = append(world.TriggerEntities, mapTile)

		if action == world.TriggerData {
			world.DataRect = r
		}

		return mapTile
	}

	createBeamEntity := func(obj *tiled.Object) gohan.Entity {
		mapTile := gohan.NewEntity()
		mapTile.AddComponent(&component.Position{
			X: obj.X,
			Y: obj.Y - 16,
		})

		sprite := &component.Sprite{
			Image: world.TileImages[obj.GID],
			Layer: world.LayerDecorationBelow,
		}
		mapTile.AddComponent(sprite)

		animateTile(obj.GID, sprite)

		return mapTile
	}

	var objects []*tiled.ObjectGroup
	var loadObjects func(grp *tiled.Group)
	loadObjects = func(grp *tiled.Group) {
		for _, subGrp := range grp.Groups {
			loadObjects(subGrp)
		}
		for _, objGrp := range grp.ObjectGroups {
			objects = append(objects, objGrp)
		}
	}
	for _, grp := range m.Groups {
		loadObjects(grp)
	}
	for _, objGrp := range m.ObjectGroups {
		objects = append(objects, objGrp)
	}

	world.Map = m
	world.ObjectGroups = objects

	var tempSpawnX, tempSpawnY float64
	for _, grp := range world.ObjectGroups {
		if !grp.Visible {
			continue
		}
		for _, obj := range grp.Objects {
			if !obj.Visible {
				continue
			}

			if grp.Name == "TRIGGERS" || grp.Name == "TEMPTRIGGERS" {
				if obj.Name == "PLAYERSPAWN" {
					world.SpawnX, world.SpawnY = obj.X, obj.Y
				} else if obj.Name == "TEMPSPAWN" {
					tempSpawnX, tempSpawnY = obj.X, obj.Y
				} else if strings.HasPrefix(obj.Name, "GET") {
					createTriggerEntity(obj)
				}
			} else if grp.Name == "BEAMS" {
				r := ObjectToRect(obj)
				beam := createBeamEntity(obj)
				shapes := addCollisionRect(r)
				world.BeamRects = append(world.BeamRects, r)
				world.BeamEntities = append(world.BeamEntities, beam)
				world.BeamShapes = append(world.BeamShapes, shapes)
			}
		}
	}
	if tempSpawnX != 0 || tempSpawnY != 0 {
		world.SpawnX, world.SpawnY = tempSpawnX, tempSpawnY
	}

	world.Player.With(func(position *component.Position) {
		position.Body.SetPosition(cp.Vector{world.SpawnX, world.SpawnY})
	})

	// Add metal collision rects.
	for y := range metalTiles {
		for x := 0; x < len(metalTiles[y]); x++ {
			collisionWidth := 0
			var offset int
			for offset = 0; offset < len(metalTiles[y])-x; offset++ {
				if !metalTiles[y][x+offset] {
					break
				}
				collisionWidth++
			}
			if collisionWidth == 0 {
				continue
			}

			r := image.Rect(x*world.TileSize, y*world.TileSize, (x+collisionWidth)*world.TileSize, (y+1)*world.TileSize)
			world.MetalRects = append(world.MetalRects, r)

			shapes := addCollisionRect(r)
			world.CollisionShapes = append(world.CollisionShapes, shapes...)

			x += offset
		}
	}
}

func ObjectToRect(o *tiled.Object) image.Rectangle {
	x, y, w, h := int(o.X), int(o.Y-16), int(16), int(32)
	return image.Rect(x, y, x+w, y+h)
}

func TileToGameCoords(x, y int) (float64, float64) {
	//return float64(x) * 32, float64(g.currentMap.Height*32) - float64(y)*32 - 32
	return float64(x) * 16, float64(y) * 16
}
